from enum import Enum

class Categories(Enum):
    function = 'function'
    gene = 'gene'
    genotype = 'genotype'
    disease = 'disease'
    phenotype = 'phenotype'
    variant = 'variant'
    genotype = 'genotype'
    model = 'model'
